from __future__ import print_function
import os
from PoserMenuLink import Link as MenuLink

try:
    import poser
except ImportError:
    from PoserLibs import POSER_FAKE as poser

TITLE = "ADP-Menu"
BUTTON_KEYWORD = "BUTTON"
TEMPDIR = "."
SCENE = poser.Scene()
UNIVERSE = SCENE.Actor("UNIVERSE")


def restore_buttons():
    """
    Back to standard button menu.
    """
    path = os.path.dirname(poser.AppLocation())
    poser.ExecFile(os.path.join(path, "Runtime", "Python", "poserScripts", "mainButtons.py"))


##############################################################################
# *** Script-template generated for each defined button
##############################################################################
SCRIPT_BODY = """### Generated Poser Button
import poser
args = {args}

if args:
    MENU.MenuObj.{funcname}(*(globals().get(a, a) for a in args))
else:
    MENU.MenuObj.{funcname}()
"""


##############################################################################
##############################################################################


class MenuClass(MenuLink):
    """
    MenuClass

    Use this class as baseclass. E.g:

        from Menu import MenuClass

        class MyMenu(MenuClass):
            def setup(self):
                ...
            def BUTTON_my_button(*args):
                ...

    """

    # Some fixed class-variables.
    __slots__ = ["_btnlist", "_tempdir", "menu_name"]

    def __init__(self, **kw):
        super(MenuClass, self).__init__()
        # <list>_btnlist holds the list of found button
        # declarations from user class.
        self._btnlist = list()
        # <str>menu_name is used to create subdirs in tempdir.
        self.menu_name = kw.pop("name", self.__class__.__name__)
        # <str>_tempdir is the place where code snippets are
        # created (Python code executed after pressing a button).
        self._tempdir = os.path.join(TEMPDIR, "SCRIPTBUTTONS")

        self.clearButtons()
        self.displayButtons()

    def setup(self):
        """
        Should be overwritten in user defined class.
        :return: None
        """
        pass

    def add(self, menuclass):
        if self.Next.__class__ != menuclass:
            MENU.MenuObj = self.Next = menuclass()
        else:
            MENU.MenuObj = self.Next
            MENU.MenuObj.clearButtons()
            MENU.MenuObj.displayButtons()

    def back(self):
        if self.Prev is None:
            print("No Prev")
        else:
            MENU.MenuObj = self.Prev
            MENU.MenuObj.clearButtons()
            MENU.MenuObj.displayButtons()

    @staticmethod
    def clearButtons():
        """
        Clear all Poser ScriptButtons.
        :return: None
        """
        for idx in range(1, 11):
            poser.DefineScriptButton(idx, "", "")

    @staticmethod
    def clearTempDir(path=TEMPDIR):
        if not os.path.exists(path):
            return

        for entry in os.listdir(path):
            p = os.path.join(path, entry)
            if os.path.isdir(p):
                MenuClass.clearTempDir(p)
                try:
                    os.remove(p)
                except Exception:
                    pass
                continue

            if entry.startswith(BUTTON_KEYWORD):
                if os.path.isfile(p):
                    try:
                        os.remove(p)
                    except Exception:
                        pass

    def _extractBtnData(self, func):
        """
        Extract button-data from method __doc__ string.
        :return: btn_nr, scriptname, args, title or name

        btn_nr = <int>, a buttons index number.
                 Indexnumber goes from 1 to 10. IndexError is raised if
                 button number is not in range.
        scriptname = <str>, absolute path-/filename (generated script to call).
                     An IOError is raised if script can't be created.
        args = <list>, array of arguments this button-function is called with.
                       Default: empty list.
                       Arguments may be a comma-seperated list.
        title or name = <str>, title/name displayed in button.
                        Default: empty string.

        Each keyword in __doc__ string must be enclosed in ":"
        (:nr:, :buttonnr:, :title:, :name:, :args:)
        Substitutes :nr: == :buttonnr:,
                    :title: == :name:.
        Keywords are case-insensitive.
        Anything after '#' is ignored in __doc__.
        E.g.:
          :nr: 2 # Script button number 2.
          :buttonnr: 2 # Same as above.
          :title: List Actors # Description for this button.
          :name: List Actors # Same as above.
          :args: a, b, c # Comma seperated list of arguments.
                           If for instance 'a' is defined global, content
                           of variable a is returned, else the string 'a'.
        """

        def extract(s):
            empty = None
            # This one does the actual work.
            if len(func.__doc__) == 0:
                return empty

            try:
                i = func.__doc__.lower().find(s.lower())
            except AttributeError:
                # Maybe no __doc__? Or no keyword?
                return empty
            if i < 0:
                return empty

            i += len(s)
            eol = func.__doc__.find("\n", i)
            ret = str(func.__doc__[i:eol].split("#", 1)[0] or "").strip()

            return ret or empty

        def nr_from_name():
            nr = func.__name__.split("_", 1)[-1].split("_", 1)[0]
            i = 0
            while i < len(nr) and nr[i].isdigit():
                i += 1
            return int(nr[:i]) if i else 0

        if not os.path.exists(self._tempdir):
            os.makedirs(self._tempdir)

        button_nr = int(extract(":nr:") or extract(":buttonnr:") or nr_from_name())
        button_ext = func.__name__.split(BUTTON_KEYWORD, 1)[-1]
        args = extract(":args:")
        #        print("FUNC: %s, ButtonNr: %s, ButtonExt: %s, Args: %s" % (func.__name__, button_nr, button_ext, args))
        args = [a.strip() for a in args.split(",")] if args else list()

        return button_nr, \
               button_ext, \
               args, \
               extract(":title:") or extract(":name:") or func.__name__.rsplit("_", 1)[-1]

    def find_button_declarations(self):
        """
        Analyse class to find declared Buttons.

        Each class method found with a name starting with predefined
        keyword ('Button' by default) is assumed to be a declaration of
        a Poser-Python Script-Button.
        The __doc__ part of these methods should have entries to
        describe this button.

        :return: None
        """
        for entry in dir(self):
            if not entry.startswith(BUTTON_KEYWORD):
                continue
            func = getattr(self, entry)
            if not callable(func):
                continue
            button_nr, func_ext, args, title = self._extractBtnData(func)
            self._btnlist.append((func, int(button_nr), args, title, func_ext))

    def displayButtons(self):
        """
        Show all defined buttons from this class.
        :return: None
        """

        def substitute_args(args, locals=locals()):
            new_args = args

            def repl(s, x):
                if s in new_args:
                    i = new_args.index(s)
                    new_args[i] = x

            repl("$index", button_nr)
            repl("$name", func.__name__)
            repl("$script", script)

            return new_args

        if len(self._btnlist) == 0:
            self.find_button_declarations()

        occupied = set()  # Index-numbers already defined.
        todo = list()  # List of buttons defined without index.
        backnext = "back next".split()  # Used for auto-index.

        self.clearTempDir(self._tempdir)

        for func, button_nr, args, title, func_ext in self._btnlist:
            # Use predefined list to actually create the buttons.
            if button_nr > 10:
                raise SyntaxError("Too much buttons defined in class '%s'." % self.__class__.__name__)

            if button_nr < 1 and func_ext in backnext:
                # Buttons with index 0 and with 'next' or 'back' in name.
                idx = backnext.index(func_ext)
                button_nr = 10 - idx
                if not title:
                    # Generate a title if none is declared.
                    title = ("<<<", ">>>")[idx] + backnext[idx].capitalize()

            if button_nr > 0:
                # Buttons with index declared or computed.
                occupied.add(button_nr)
                script = os.path.abspath(os.path.join(self._tempdir, "BUTTON_%02d.py" % button_nr))
                poser.DefineScriptButton(button_nr, script, title)
                with open(script, "w") as fh:
                    print(SCRIPT_BODY.format(args=substitute_args(args), funcname=func.__name__), file=fh)
            else:
                # Buttons without index are stored to be processed later.
                todo.append((args, func, title))

        if todo:
            # Process buttons without index.
            for button_nr in range(1, 11):
                if button_nr not in occupied:
                    args, func, title = todo.pop(0)
                    script = os.path.abspath(os.path.join(self._tempdir, "BUTTON_%02d.py" % button_nr))
                    poser.DefineScriptButton(button_nr, script, title)
                    with open(script, "w") as fh:
                        print(SCRIPT_BODY.format(args=substitute_args(args), funcname=func.__name__), file=fh)
                    if len(todo) == 0:
                        # all set.
                        break
            if todo:
                # still some buttons left?
                s = ",".join(t[1] for t in todo)
                raise SyntaxError("No more room for defined buttons [%s]" % s)

        # Make sure generated script can find us.
        MENU.MenuObj = self

    def rename(self, index, newTitle):
        for idx, entry in enumerate(self._btnlist):
            func, button_nr, script, args, title, func_ext = entry
            if button_nr == index:
                self._btnlist[idx] = func, button_nr, script, args, newTitle, func_ext
                poser.DefineScriptButton(button_nr, script, newTitle)

    def __del__(self):
        # make sure back references are deleted.
        try:
            if MENU.MenuObj == self:
                MENU.MenuObj = None
        except NameError:
            pass
        except Exception:
            pass
        self.clearTempDir(self._tempdir)


##############################################################################
# *** Try to make global available variable MENU as resistent as possible
# *** if it not yet exists.
# *** This works by defining MENU as a builtin function.
##############################################################################

MENU = globals().get("MENU", None)
if MENU.__class__.__name__ != "__MENU__":
    class __MENU__(object):
        __slots__ = ["MenuObj", "MenuClass", "MenuList"]

        def __init__(self):
            super(__MENU__, self).__init__()
            self.MenuClass = MenuClass
            self.MenuObj = None

    MENU = __MENU__()
try:
    __builtins__.MENU = MENU
except AttributeError:
    __builtins__["MENU"] = MENU

class MenuError(Exception):
    pass

MENU.MenuClass = MenuClass
